/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.spark.impl.common.repartition;

import lombok.extern.slf4j.Slf4j;
import org.apache.spark.Partitioner;

import java.util.Random;

@Slf4j
public class BalancedPartitioner extends Partitioner {
    private final int numPartitions; //Total number of partitions
    private final int elementsPerPartition;
    private final int remainder;
    private Random r;

    public BalancedPartitioner(int numPartitions, int elementsPerPartition, int remainder) {
        this.numPartitions = numPartitions;
        this.elementsPerPartition = elementsPerPartition;
        this.remainder = remainder;
    }

    @Override
    public int numPartitions() {
        return numPartitions;
    }

    @Override
    public int getPartition(Object key) {
        int elementIdx = key.hashCode();

        //First 'remainder' executors get elementsPerPartition+1 each; the remainder get
        // elementsPerPartition each. This is because the total number of examples might not be an exact multiple
        // of the number of cores in the cluster

        //Work out: which partition it belongs to...
        if (elementIdx <= (elementsPerPartition + 1) * remainder) {
            //This goes into one of the larger partitions (of size elementsPerPartition+1)
            int outputPartition = elementIdx / (elementsPerPartition + 1);
            if (outputPartition >= numPartitions) {
                //Should never happen, unless there's some up-stream problem with calculating elementsPerPartition
                outputPartition = getRandom().nextInt(numPartitions);
                log.trace("Random partition assigned (1): elementIdx={}, numPartitions={}, elementsPerPartition={}, remainder={}",
                        elementIdx, numPartitions, elementsPerPartition, remainder);
            }
            return outputPartition;
        } else {
            //This goes into one of the standard size partitions (of size elementsPerPartition)
            int numValsInLargerPartitions = remainder * (elementsPerPartition + 1);
            int idxInSmallerPartitions = elementIdx - numValsInLargerPartitions;
            int smallPartitionIdx = idxInSmallerPartitions / elementsPerPartition;
            int outputPartition = remainder + smallPartitionIdx;
            if (outputPartition >= numPartitions) {
                //Should never happen, unless there's some up-stream problem with calculating elementsPerPartition
                outputPartition = getRandom().nextInt(numPartitions);
                log.trace("Random partition assigned (2): elementIdx={}, numPartitions={}, elementsPerPartition={}, remainder={}",
                        elementIdx, numPartitions, elementsPerPartition, remainder);
            }
            return outputPartition;
        }
    }

    private synchronized Random getRandom() {
        if (r == null)
            r = new Random();
        return r;
    }
}
