/*
 *
 *    ******************************************************************************
 *    *
 *    * This program and the accompanying materials are made available under the
 *    * terms of the Apache License, Version 2.0 which is available at
 *    * https://www.apache.org/licenses/LICENSE-2.0.
 *    *
 *    *  See the NOTICE file distributed with this work for additional
 *    *  information regarding copyright ownership.
 *    * Unless required by applicable law or agreed to in writing, software
 *    * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *    * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *    * License for the specific language governing permissions and limitations
 *    * under the License.
 *    *
 *    * SPDX-License-Identifier: Apache-2.0
 *    *****************************************************************************
 *
 */

package net.brutex.ai.dnn.networks;

import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import lombok.Getter;
import lombok.NonNull;
import lombok.Setter;
import net.brutex.ai.dnn.api.IModel;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.gradient.Gradient;
import org.nd4j.linalg.api.ndarray.INDArray;


/**
 * Artificial Neural Network An artificial neural network (1) takes some input data, and (2)
 * transforms this input data by calculating a weighted sum over the inputs and (3) applies a
 * non-linear function to this transformation to calculate an intermediate state. The three steps
 * above constitute what is known as a layer, and the transformative function is often referred to
 * as a unit. The intermediate states—often termed features—are used as the input into another
 * layer.
 * <p>
 * Through repetition of these steps, the artificial neural network learns multiple layers of
 * non-linear features, which it then combines in a final layer to create a prediction.
 * <p>
 * The neural network learns by generating an error signal that measures the difference between the
 * predictions of the network and the desired values and then using this error signal to change the
 * weights (or parameters) so that predictions get more accurate.
 */
public abstract class ArtificialNeuralNetwork implements IModel {

  /**
   * A neural network is created from a configuration.
   *
   * @param conf The (new net.brutex.ai) configuration for the network
   */
  @Getter
  @Setter
  @NonNull
  private NeuralNetConfiguration netConfiguration;

  @Getter
  @Setter
  private Map<String, INDArray> paramTable;

  /**
   * Table of parameters by key, for backprop. For many models (dense layers, etc) - all parameters
   * are backprop parameters
   *
   * @param backpropParamsOnly If true, return backprop params only. If false: return all params
   *                           (equivalent to paramsTable())
   */
  @Override
  public Map<String, INDArray> getParamTable(boolean backpropParamsOnly) {
    return paramTable;
  }


  /**
   * Set the parameters of the network. Note that the parameter keys must match the format as
   * described in {@link #getParam(String)} and {@link #getParamTable()}. Note that the values of the
   * parameters used as an argument to this method are copied - i.e., it is safe to later
   * modify/reuse the values in the provided paramTable without this impacting the network.
   *
   * @param paramTable Parameters to set
   */
  @Override
  public void setParamTable(Map<String, INDArray> paramTable) {
    Map<String, INDArray> currParamTable = getParamTable();
    if(currParamTable == null) {
      currParamTable = paramTable;
    } else if (!currParamTable.keySet().equals(paramTable.keySet())) {
      throw new IllegalArgumentException(
              "Cannot set param table: parameter keys do not match.\n" + "Current: "
                      + currParamTable.keySet() + "\nTo set: " + paramTable.keySet());
    }

    for (String s : paramTable.keySet()) {
      INDArray curr = currParamTable.get(s);
      INDArray toSet = paramTable.get(s);
      if (!Arrays.equals(curr.shape(), toSet.shape())) {
        throw new IllegalArgumentException(
                "Cannot set parameter table: parameter \"" + s + "\" shapes "
                        + "do not match. Current = " + Arrays.toString(curr.shape()) + ", to set = "
                        + Arrays.toString(toSet.shape()));
      }
    }

    //Now that we've checked ALL params (to avoid leaving net in half-modified state)
    for (String s : paramTable.keySet()) {
      INDArray curr = currParamTable.get(s);
      INDArray toSet = paramTable.get(s);
      curr.assign(toSet);
    }
  }



  /**
   * Create a new network from configuration
   *
   * @param conf the configuration
   */
  public ArtificialNeuralNetwork(NeuralNetConfiguration conf) {
    this.netConfiguration = conf;
  }

  /**
   * Update all parameters (for all parameter types) with the given gradient.
   *
   * @param gradient the gradients to add
   */
  public void update(Gradient gradient) {
    for (String paramType : gradient.gradientForVariable().keySet()) {
      update(gradient.getGradientFor(paramType), paramType);
    }
  }

  /**
   * Update the parameters of a given type with a given gradient.
   *
   * @param gradient  the gradient to apply
   * @param paramType
   */
  public void update(INDArray gradient, String paramType) {
    setParam(paramType, getParam(paramType).addi(gradient));
  }




}
