/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.nn.conf;

import com.fasterxml.jackson.annotation.JsonIdentityInfo;
import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonIgnoreProperties;
import com.fasterxml.jackson.annotation.ObjectIdGenerators;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.exc.InvalidTypeIdException;
import com.fasterxml.jackson.databind.node.ArrayNode;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import lombok.*;
import lombok.experimental.SuperBuilder;
import lombok.extern.jackson.Jacksonized;
import lombok.extern.slf4j.Slf4j;
import net.brutex.ai.dnn.api.IModel;
import org.deeplearning4j.nn.api.OptimizationAlgorithm;
import org.deeplearning4j.nn.api.layers.LayerConstraint;
import org.deeplearning4j.nn.conf.distribution.Distribution;
import org.deeplearning4j.nn.conf.dropout.Dropout;
import org.deeplearning4j.nn.conf.dropout.IDropout;
import org.deeplearning4j.nn.conf.inputs.InputType;
import org.deeplearning4j.nn.conf.layers.BaseLayerConfiguration;
import org.deeplearning4j.nn.conf.layers.BaseOutputLayer;
import org.deeplearning4j.nn.conf.layers.BaseRecurrentLayer;
import org.deeplearning4j.nn.conf.layers.Convolution1DLayer;
import org.deeplearning4j.nn.conf.layers.ConvolutionLayer;
import org.deeplearning4j.nn.conf.layers.DenseLayer;
import org.deeplearning4j.nn.conf.layers.EmbeddingLayer;
import org.deeplearning4j.nn.conf.layers.FeedForwardLayer;
import org.deeplearning4j.nn.conf.layers.GlobalPoolingLayer;
import org.deeplearning4j.nn.conf.layers.LayerConfiguration;
import org.deeplearning4j.nn.conf.layers.OutputLayer;
import org.deeplearning4j.nn.conf.layers.recurrent.LastTimeStep;
import org.deeplearning4j.nn.conf.memory.LayerMemoryReport;
import org.deeplearning4j.nn.conf.memory.MemoryReport;
import org.deeplearning4j.nn.conf.memory.NetworkMemoryReport;
import org.deeplearning4j.nn.conf.serde.JsonMappers;
import org.deeplearning4j.nn.conf.stepfunctions.DefaultStepFunction;
import org.deeplearning4j.nn.conf.stepfunctions.StepFunction;
import org.deeplearning4j.nn.conf.weightnoise.IWeightNoise;
import org.deeplearning4j.nn.weights.IWeightInit;
import org.deeplearning4j.nn.weights.WeightInit;
import org.deeplearning4j.util.OutputLayerUtil;
import org.nd4j.linalg.activations.Activation;
import org.nd4j.linalg.activations.IActivation;
import org.nd4j.linalg.activations.impl.ActivationSigmoid;
import org.nd4j.linalg.api.buffer.DataType;
import org.nd4j.linalg.factory.Nd4j;
import org.nd4j.linalg.learning.config.IUpdater;
import org.nd4j.linalg.learning.config.Sgd;
import org.nd4j.linalg.learning.regularization.Regularization;
import org.nd4j.linalg.learning.regularization.WeightDecay;
import org.nd4j.linalg.lossfunctions.LossFunctions;
import org.nd4j.linalg.lossfunctions.impl.LossBinaryXENT;
import org.nd4j.linalg.lossfunctions.impl.LossMCXENT;
import org.nd4j.linalg.lossfunctions.impl.LossMSE;
import org.nd4j.linalg.lossfunctions.impl.LossNegativeLogLikelihood;

/**
 * Deeplearning4j is a domain-specific language to configure deep neural networks, which are made of
 * multiple layers. Everything starts with a NeuralNetConfiguration, which organizes those layers
 * and their hyperparameters. Hyperparameters are variables that determine how a neural network
 * learns. They include how many times to update the weights of the model, how to initialize those
 * weights, which activation function to attach to the nodes, which optimization algorithm to use,
 * and how fast the model should learn. This is what one configuration would look like:
 * <br/><br/>
 *
 * NeuralNetConfiguration conf = NeuralNetConfiguration.builder()<br/>
 * .weightInit(WeightInit.XAVIER) .activation(Activation.RELU)<br/>
 * .optimizationAlgo(OptimizationAlgorithm.STOCHASTIC_GRADIENT_DESCENT)<br/>
 * .updater(new Sgd(0.05)) //... other hyperparameters <br/>
 *  .backprop(true)<br/>
 * .build();<br/><br/>
 *
 * With Deeplearning4j, you add a layer
 * by calling layer on the NeuralNetConfiguration.NeuralNetConfigurationBuilder(), specifying its place in the order of
 * layers (the zero-indexed layer below is the input layer), the number of input and output nodes,
 * nIn and nOut, as well as the type: DenseLayer.<br/><br/>
 *
 * .layer(0, new DenseLayer.Builder().nIn(784).nOut(250)<br/>
 * .build())<br/><br/>
 *
 * Once you've configured your net, you train the
 * model with model.fit.
 */

@Data
@Slf4j
@EqualsAndHashCode(exclude = {"iterationCount", "epochCount"})
@Jacksonized
@JsonIgnoreProperties(ignoreUnknown = true)
@JsonIdentityInfo(generator= ObjectIdGenerators.IntSequenceGenerator.class, property="@id")

//The inner builder, that we can then extend ...
@SuperBuilder //TODO fix access
public class NeuralNetConfiguration extends NeuralNetBaseBuilderConfiguration {

  private IModel net;
  private static final int DEFAULT_TBPTT_LENGTH = 20;
  private boolean initCalled = false;

  /**
   * Set constraints to be applied to all layers. Default: no constraints.<br>
   * Constraints can be used to enforce certain conditions (non-negativity of parameters, max-norm regularization,
   * etc). These constraints are applied at each iteration, after the parameters have been updated.<br>
   * Note: values set by this method will be applied to all applicable layers in the network, unless a different
   * value is explicitly set on a given layer. In other words: values set via this method are used as the default
   * value, and can be overridden on a per-layer basis.
   *
   * @param constraints Constraints to apply to all bias parameters of all layers
   */
  @lombok.Builder.Default
  protected final List<LayerConstraint> biasConstraints = new ArrayList<>();
  /**
   * Set constraints to be applied to all layers. Default: no constraints.<br>
   * Constraints can be used to enforce certain conditions (non-negativity of parameters, max-norm regularization,
   * etc). These constraints are applied at each iteration, after the parameters have been updated.<br>
   * Note: values set by this method will be applied to all applicable layers in the network, unless a different
   * value is explicitly set on a given layer. In other words: values set via this method are used as the default
   * value, and can be overridden on a per-layer basis.
   *
   * @param constraints Constraints to apply to all parameters of all layers
   */
  @lombok.Builder.Default
  protected final List<LayerConstraint> allParamContraints = new ArrayList<>();

  @Getter
  @Setter
  @NonNull
  @lombok.Builder.Default
  @Deprecated
  protected WorkspaceMode trainingWorkspaceMode = WorkspaceMode.ENABLED;
  @Getter
  @Setter
  @NonNull
  @lombok.Builder.Default
  @Deprecated
  protected WorkspaceMode inferenceWorkspaceMode = WorkspaceMode.ENABLED;
  /**
   * The type of backprop. Default setting is used for most networks (MLP, CNN etc), but optionally
   * truncated BPTT can be used for training recurrent neural networks. If using TruncatedBPTT make
   * sure you set both tBPTTForwardLength() and tBPTTBackwardLength()
   */
  @Getter
  @Setter
  @NonNull
  @lombok.Builder.Default
  protected BackpropType backpropType = BackpropType.Standard;
  /**
   * When doing truncated BPTT: how many steps of forward pass should we do before doing (truncated)
   * backprop?<br> Only applicable when doing backpropType(BackpropType.TruncatedBPTT)<br> Typically
   * tBPTTForwardLength parameter is same as the tBPTTBackwardLength parameter, but may be larger
   * than it in some circumstances (but never smaller)<br> Ideally your training data time series
   * length should be divisible by this This is the k1 parameter on pg23 of <a
   * href="http://www.cs.utoronto.ca/~ilya/pubs/ilya_sutskever_phd_thesis.pdf">http://www.cs.utoronto.ca/~ilya/pubs/ilya_sutskever_phd_thesis.pdf</a>
   *
   * @param forwardLength Forward length > 0, >= backwardLength
   */
  @Getter
  @Setter
  @lombok.Builder.Default
  protected int tbpttFwdLength = 20;
  /**
   * When doing truncated BPTT: how many steps of backward should we do?<br> Only applicable when
   * doing backpropType(BackpropType.TruncatedBPTT)<br> This is the k2 parameter on pg23 of <a
   * href="http://www.cs.utoronto.ca/~ilya/pubs/ilya_sutskever_phd_thesis.pdf">http://www.cs.utoronto.ca/~ilya/pubs/ilya_sutskever_phd_thesis.pdf</a>
   *
   * @param backwardLength <= forwardLength
   */
  @Getter
  @Setter
  @lombok.Builder.Default
  protected int tbpttBackLength = 20;
  /**
   * Creates and returns a copy of this object.
   *
   * @return a clone of this instance.
   * @throws CloneNotSupportedException if the object's class does not support the {@code Cloneable}
   * interface. Subclasses that override the {@code clone} method can also throw this exception to
   * indicate that an instance cannot be cloned.
   * @see Cloneable
   */

  //Nd4j.getRandom().setSeed(getConf(0).getSeed()); //TODO
  //Counter for the number of parameter updates so far
  // This is important for learning rate schedules, for example, and is stored here to ensure it is persisted
  // for Spark and model serialization
  @Getter
  @Setter
  @lombok.Builder.Default
  protected int iterationCount = 0;
  //Counter for the number of epochs completed so far. Used for per-epoch schedules
  @Getter
  @Setter
  @lombok.Builder.Default
  protected int epochCount = 0;
  @lombok.Builder.Default
  protected double dampingFactor = 100;
  //gradient keys used for ensuring order when getting and setting the gradient
  @lombok.Builder.Default
  protected List<String> netWideVariables = new ArrayList<>();
  @Getter
  @Setter
  @lombok.Builder.Default
  private boolean miniBatch = false;
  /**
   * A seed for this network, will be random if not specified.

  @Getter
  @Setter
  @lombok.Builder.Default
  private long seed = new Random().nextLong(); */
  /**
   * The default {@link CacheMode} for this configuration. Will be set to "NONE" if not specified
   * otherwise. This method defines how/if preOutput cache is handled: NONE: cache disabled (default
   * value) HOST: Host memory will be used DEVICE: GPU memory will be used (on CPU backends effect
   * will be the same as for HOST)
   * <p>
   * Valid values are<br/> CacheMode.NONE,<br/> CacheMode.HOST or<br/> CacheMode.DEVICE<br/>
   *
   * @param cacheMode
   */
  @NonNull
  @Getter
  @Setter
  @lombok.Builder.Default
  private CacheMode cacheMode = CacheMode.NONE;
  /**
   * The list of layer configurations in this configuration. They will be indexed automatically as
   * the layers get added starting with index 0.
   */

  @lombok.Builder.Default
  @Getter
  private String name = "Anonymous INeuralNetworkConfiguration";
  /**
   * The {@link InputType} of the data for this network configuration
   */
  @Getter
  @Setter
  private InputType inputType;
  /**
   * Set the DataType for the network parameters and activations for all layers in the network.
   * Default: Float
   *
   * @param dataType Datatype to use for parameters and activations
   */
  @Getter
  @Setter
  @lombok.Builder.Default
  @NonNull
  private DataType dataType = DataType.FLOAT;
  /**
   * Whether to override the nIn configuration forcibly upon construction. Default value is true.
   *
   * @return builder pattern
   */
  @Getter
  @Setter
  @lombok.Builder.Default
  private boolean overrideNinUponBuild = true;
  /**
   * Enabled by default. If enabled, the output layer configuration will be validated, to throw an
   * exception on likely invalid outputs - such as softmax + nOut=1, or LossMCXENT + Tanh.<br> If
   * disabled (false) no output layer validation will be performed.<br> Disabling this validation is
   * not recommended, as the configurations that fail validation usually will not be able to learn
   * correctly. However, the option to disable this validation is provided for advanced users when
   * creating non-standard architectures.
   *
   * @param validate If true: validate output layer configuration. False: don't validate
   */
  @Getter
  @Setter
  @lombok.Builder.Default
  private boolean validateOutputLayerConfig = true;
  /**
   * Enabled by default. If enabled, an exception will be throw when using the (invalid) combination
   * of truncated backpropagation through time (TBPTT) with either a GlobalPoolingLayer or
   * LastTimeStepLayer.<br> It is possible to disable this validation to allow what is almost
   * certainly an invalid configuration to be used, however this is not recommended.
   *
   * @param validate Whether TBPTT validation should be performed
   */
  @Getter
  @Setter
  @lombok.Builder.Default
  private boolean validateTbpttConfig = true;
  /**
   * Gradient updater configuration. For example, {@link org.nd4j.linalg.learning.config.Adam} or
   * {@link org.nd4j.linalg.learning.config.Nesterovs}<br>
   * Note: values set by this method will be applied to all applicable layers in the network, unless
   * a different value is explicitly set on a given layer. In other words: values set via this
   * method are used as the default value, and can be overridden on a per-layer basis.
   *
   * @param updater Updater to use
   */
  @Getter @Setter @Builder.Default private IUpdater updater = new Sgd();
  /**
   * Gradient normalization strategy. Used to specify gradient renormalization, gradient clipping
   * etc. See {@link GradientNormalization} for details<br> Note: values set by this method will be
   * applied to all applicable layers in the network, unless a different value is explicitly set on
   * a given layer. In other words: values set via this method are used as the default value, and
   * can be overridden on a per-layer basis.
   *
   * @param gradientNormalization Type of normalization to use. Defaults to None.
   * @see GradientNormalization
   */
  @Getter
  @Setter
  @NonNull
  @lombok.Builder.Default
  private GradientNormalization gradientNormalization = GradientNormalization.None;
  /**
   * Threshold for gradient normalization, only used for GradientNormalization.ClipL2PerLayer,
   * GradientNormalization.ClipL2PerParamType, and
   * GradientNormalization.ClipElementWiseAbsoluteValue<br> Not used otherwise.<br> L2 threshold for
   * first two types of clipping, or absolute value threshold for last type of clipping.<br> Note:
   * values set by this method will be applied to all applicable layers in the network, unless a
   * different value is explicitly set on a given layer. In other words: values set via this method
   * are used as the default value, and can be overridden on a per-layer basis.
   */
  @Getter
  @Setter
  private double gradientNormalizationThreshold;

  // whether to constrain the gradient to unit norm or not
  @Getter @Setter @Builder.Default private StepFunction stepFunction = new DefaultStepFunction();

  @Getter
  @Setter
  @lombok.Builder.Default
  private OptimizationAlgorithm optimizationAlgo = OptimizationAlgorithm.STOCHASTIC_GRADIENT_DESCENT;
  @Getter
  @Setter
  @lombok.Builder.Default
  private int maxNumLineSearchIterations = 5;
  /**
   * Set the regularization for the parameters (excluding biases) - for example {@link WeightDecay}<br>
   * Note: values set by this method will be applied to all applicable layers in the network, unless a different
   * value is explicitly set on a given layer. In other words: values set via this method are used as the default
   * value, and can be overridden on a per-layer basis.<br>
   *
   * @param regularization Regularization to apply for the network parameters/weights (excluding biases)
   */
  @Getter
  @lombok.Builder.Default
  private List<Regularization> regularization = new ArrayList<>();
  /**
   * Set the regularization for the biases only - for example {@link WeightDecay}<br>
   * Note: values set by this method will be applied to all applicable layers in the network, unless a different
   * value is explicitly set on a given layer. In other words: values set via this method are used as the default
   * value, and can be overridden on a per-layer basis.<br>
   *
   * @param regularizationBias Regularization to apply for the network biases only
   */
  @Getter
  @lombok.Builder.Default
  private List<Regularization> regularizationBias = new ArrayList<>();

  /**
   * Gradient updater configuration, for the biases only. If not set, biases will use the updater as
   * set by {@link #setUpdater(IUpdater)}<br>
   * Note: values set by this method will be applied to all applicable layers in the network, unless a different
   * value is explicitly set on a given layer. In other words: values set via this method are used as the default
   * value, and can be overridden on a per-layer basis.
   *
   * @param updater Updater to use for bias parameters
   */
  @Getter
  @Setter
  @lombok.Builder.Default
  private IUpdater biasUpdater = null;
  @Getter
  @Setter
  @lombok.Builder.Default
  private IActivation activation = new ActivationSigmoid();

  /**
   * Sets the convolution mode for convolutional layers, which impacts padding and output sizes.
   * See {@link ConvolutionMode} for details. Defaults to ConvolutionMode.TRUNCATE<br>
   * Note: values set by this method will be applied to all applicable layers in the network, unless a different
   * value is explicitly set on a given layer. In other words: values set via this method are used as the default
   * value, and can be overridden on a per-layer basis.
   * @param convolutionMode Convolution mode to use
   */
  @Getter
  @Setter
  @lombok.Builder.Default
  private ConvolutionMode convolutionMode = ConvolutionMode.Truncate;
  /**
   * Sets the cuDNN algo mode for convolutional layers, which impacts performance and memory usage of cuDNN.
   * See {@link ConvolutionLayer.AlgoMode} for details.  Defaults to "PREFER_FASTEST", but "NO_WORKSPACE" uses less memory.
   * <br>
   * Note: values set by this method will be applied to all applicable layers in the network, unless a different
   * value is explicitly set on a given layer. In other words: values set via this method are used as the default
   * value, and can be overridden on a per-layer basis.
   * @param cudnnAlgoMode cuDNN algo mode to use
   */
  @Getter
  @Setter
  @lombok.Builder.Default
  private ConvolutionLayer.AlgoMode cudnnAlgoMode = ConvolutionLayer.AlgoMode.PREFER_FASTEST;
  @Getter
  @Setter
  @lombok.Builder.Default
  private boolean minimize = true;
  /**
   * Set the dropout for all layers in this network<br>
   * Note: values set by this method will be applied to all applicable layers in the network, unless a different
   * value is explicitly set on a given layer. In other words: values set via this method are used as the default
   * value, and can be overridden on a per-layer basis.
   *          * Dropout probability. This is the probability of <it>retaining</it> each input activation value for a layer.
   *          * dropOut(x) will keep an input activation with probability x, and set to 0 with probability 1-x.<br>
   *          * dropOut(0.0) is a special value / special case - when set to 0.0., dropout is disabled (not applied). Note
   *          * that a dropout value of 1.0 is functionally equivalent to no dropout: i.e., 100% probability of retaining
   *          * each input activation.<br>
   *          * <p>
   *          * Note 1: Dropout is applied at training time only - and is automatically not applied at test time
   *          * (for evaluation, etc)<br>
   *          * Note 2: This sets the probability per-layer. Care should be taken when setting lower values for
   *          * complex networks (too much information may be lost with aggressive (very low) dropout values).<br>
   *          * Note 3: Frequently, dropout is not applied to (or, has higher retain probability for) input (first layer)
   *          * layers. Dropout is also often not applied to output layers. This needs to be handled MANUALLY by the user
   *          * - set .dropout(0) on those layers when using global dropout setting.<br>
   *          * Note 4: Implementation detail (most users can ignore): DL4J uses inverted dropout, as described here:
   *          * <a href="http://cs231n.github.io/neural-networks-2/">http://cs231n.github.io/neural-networks-2/</a>
   *          * </p>
   *          *<br>
   *          * Note: values set by this method will be applied to all applicable layers in the network, unless a different
   *          * value is explicitly set on a given layer. In other words: values set via this method are used as the default
   *          * value, and can be overridden on a per-layer basis.
   *          *
   *          * @param inputRetainProbability Dropout probability (probability of retaining each input activation value for a layer)
   *          * @see #dropOut(IDropout)
   *
   *
   * @param dropout Dropout, such as {@link Dropout}, {@link org.deeplearning4j.nn.conf.dropout.GaussianDropout},
   *                {@link org.deeplearning4j.nn.conf.dropout.GaussianNoise} etc
   * @return
   */
  @Getter
  @Setter
  private IDropout idropOut;
  /**
   * Set the weight noise (such as {@link org.deeplearning4j.nn.conf.weightnoise.DropConnect} and
   * {@link org.deeplearning4j.nn.conf.weightnoise.WeightNoise}) for the layers in this network.<br>
   * Note: values set by this method will be applied to all applicable layers in the network, unless a different
   * value is explicitly set on a given layer. In other words: values set via this method are used as the default
   * value, and can be overridden on a per-layer basis.
   *
   * @param weightNoise Weight noise instance to use
   */
  @Getter
  @Setter
  private IWeightNoise weightNoise;
  @Getter
  @Setter
  @lombok.Builder.Default
  private double biasInit = 0.0;
  @Getter
  @Setter
  @lombok.Builder.Default
  private double gainInit = 1.0;

  /**
   * Create a neural net configuration from json
   *
   * @param json the neural net configuration from json
   * @return {@link NeuralNetConfiguration}
   */
  public static NeuralNetConfiguration fromJson(String json) {
    NeuralNetConfiguration conf;
    ObjectMapper mapper = NeuralNetConfiguration.mapper();
    try {
      conf = mapper.readValue(json, NeuralNetConfiguration.class);
    } catch (InvalidTypeIdException e) {
      if (e.getMessage().contains("@class")) {
        try {
          //JSON may be legacy (1.0.0-alpha or earlier), attempt to load it using old format
          return JsonMappers.getLegacyMapper().readValue(json, NeuralNetConfiguration.class);
        } catch (InvalidTypeIdException e2) {
          //Check for legacy custom layers: "Could not resolve type id 'CustomLayer' as a subtype of [simple type, class org.deeplearning4j.nn.conf.layers.ILayer]: known type ids = [Bidirectional, CenterLossOutputLayer, CnnLossLayer, ..."
          //1.0.0-beta5: dropping support for custom layers defined in pre-1.0.0-beta format. Built-in layers from these formats still work
          String msg = e2.getMessage();
          if (msg != null && msg.contains("Could not resolve type id")) {
            throw new RuntimeException(
                "Error deserializing NeuralNetConfiguration - configuration may have a custom " +
                    "layer, vertex or preprocessor, in pre version 1.0.0-beta JSON format.\nModels in legacy format with custom"
                    +
                    " layers should be loaded in 1.0.0-beta to 1.0.0-beta4 and saved again, before loading in the current version of DL4J",
                e);
          }
          throw new RuntimeException(e2);
        } catch (IOException e2) {
          throw new RuntimeException(e2);
        }
      }
      throw new RuntimeException(e);
    } catch (IOException e) {
      //Check if this exception came from legacy deserializer...
      String msg = e.getMessage();
      if (msg != null && msg.contains("legacy")) {
        throw new RuntimeException(
            "Error deserializing NeuralNetConfiguration - configuration may have a custom " +
                "layer, vertex or preprocessor, in pre version 1.0.0-alpha JSON format. These layers can be "
                +
                "deserialized by first registering them with NeuralNetConfiguration.registerLegacyCustomClassesForJSON(Class...)",
            e);
      }
      throw new RuntimeException(e);
    }

    //To maintain backward compatibility after loss function refactoring (configs generated with v0.5.0 or earlier)
    // Previously: enumeration used for loss functions. Now: use classes
    // IN the past, could have only been an OutputLayer or RnnOutputLayer using these enums
    int layerCount = 0;
    JsonNode confs = null;
    for (LayerConfiguration nnc : conf.getFlattenedLayerConfigurations()) {
      LayerConfiguration l = nnc;
      if (l instanceof BaseOutputLayer && ((BaseOutputLayer) l).getLossFn() == null) {
        //lossFn field null -> may be an old config format, with lossFunction field being for the enum
        //if so, try walking the JSON graph to extract out the appropriate enum value

        BaseOutputLayer ol = (BaseOutputLayer) l;
        try {
          JsonNode jsonNode = mapper.readTree(json);
          if (confs == null) {
            confs = jsonNode.get("confs");
          }
          if (confs instanceof ArrayNode) {
            ArrayNode layerConfs = (ArrayNode) confs;
            JsonNode outputLayerNNCNode = layerConfs.get(layerCount);
            if (outputLayerNNCNode == null) {
              throw new RuntimeException(
                  "should never happen"); //return conf; //Should never happen...
            }
            JsonNode outputLayerNode = outputLayerNNCNode.get("layer");

            JsonNode lossFunctionNode = null;
            if (outputLayerNode.has("output")) {
              lossFunctionNode = outputLayerNode.get("output").get("lossFunction");
            } else if (outputLayerNode.has("rnnoutput")) {
              lossFunctionNode = outputLayerNode.get("rnnoutput").get("lossFunction");
            }

            if (lossFunctionNode != null) {
              String lossFunctionEnumStr = lossFunctionNode.asText();
              LossFunctions.LossFunction lossFunction = null;
              try {
                lossFunction = LossFunctions.LossFunction.valueOf(lossFunctionEnumStr);
              } catch (Exception e) {
                log.warn(
                    "OutputLayer with null LossFunction or pre-0.6.0 loss function configuration detected: could not parse JSON",
                    e);
              }

              if (lossFunction != null) {
                switch (lossFunction) {
                  case MSE:
                    ol.setLossFn(new LossMSE());
                    break;
                  case XENT:
                    ol.setLossFn(new LossBinaryXENT());
                    break;
                  case NEGATIVELOGLIKELIHOOD:
                    ol.setLossFn(new LossNegativeLogLikelihood());
                    break;
                  case MCXENT:
                    ol.setLossFn(new LossMCXENT());
                    break;

                  //Remaining: TODO
                  case SQUARED_LOSS:
                  case RECONSTRUCTION_CROSSENTROPY:
                  default:
                    log.warn(
                        "OutputLayer with null LossFunction or pre-0.6.0 loss function configuration detected: could not set loss function for {}",
                        lossFunction);
                    break;
                }
              }
            }

          } else {
            log.warn(
                "OutputLayer with null LossFunction or pre-0.6.0 loss function configuration detected: could not parse JSON: layer 'confs' field is not an ArrayNode (is: {})",
                (confs != null ? confs.getClass() : null));
          }
        } catch (IOException e) {
          log.warn(
              "OutputLayer with null LossFunction or pre-0.6.0 loss function configuration detected: could not parse JSON",
              e);
          break;
        }
      }

      //Also, pre 0.7.2: activation functions were Strings ("activationFunction" field), not classes ("activationFn")
      //Try to load the old format if necessary, and create the appropriate IActivation instance
      if ((l instanceof BaseLayerConfiguration) && ((BaseLayerConfiguration) l).getActivationFn() == null) {
        try {
          JsonNode jsonNode = mapper.readTree(json);
          if (confs == null) {
            confs = jsonNode.get("confs");
          }
          if (confs instanceof ArrayNode) {
            ArrayNode layerConfs = (ArrayNode) confs;
            JsonNode outputLayerNNCNode = layerConfs.get(layerCount);
            if (outputLayerNNCNode == null) {
              throw new RuntimeException(
                  "Should never happen"); //return conf; //Should never happen...
            }
            JsonNode layerWrapperNode = outputLayerNNCNode.get("layer");

            if (layerWrapperNode == null || layerWrapperNode.size() != 1) {
              continue;
            }

            JsonNode layerNode = layerWrapperNode.elements().next();
            JsonNode activationFunction = layerNode.get(
                "activationFunction"); //Should only have 1 element: "dense", "output", etc

            if (activationFunction != null) {
              IActivation ia = Activation.fromString(activationFunction.asText())
                  .getActivationFunction();
              ((BaseLayerConfiguration) l).setActivationFn(ia);
            }
          }

        } catch (IOException e) {
          log.warn(
              "ILayer with null ActivationFn field or pre-0.7.2 activation function detected: could not parse JSON",
              e);
        }
      }

      if (!handleLegacyWeightInitFromJson(json, l, mapper, confs, layerCount)) {
        return conf;
      }

      layerCount++;
    }
    return conf;
  }

  /**
   * Handle {@link WeightInit} and {@link Distribution} from legacy configs in Json format. Copied
   * from handling of {@link Activation} above.
   *
   * @return True if all is well and layer iteration shall continue. False else-wise.
   */
  private static boolean handleLegacyWeightInitFromJson(String json, LayerConfiguration l,
      ObjectMapper mapper,
      JsonNode confs, int layerCount) {
    if ((l instanceof BaseLayerConfiguration) && ((BaseLayerConfiguration) l).getWeightInit() == null) {
      try {
        JsonNode jsonNode = mapper.readTree(json);
        if (confs == null) {
          confs = jsonNode.get("confs");
        }
        if (confs instanceof ArrayNode) {
          ArrayNode layerConfs = (ArrayNode) confs;
          JsonNode outputLayerNNCNode = layerConfs.get(layerCount);
          if (outputLayerNNCNode == null) {
            return false; //Should never happen...
          }
          JsonNode layerWrapperNode = outputLayerNNCNode.get("layer");

          if (layerWrapperNode == null || layerWrapperNode.size() != 1) {
            return true;
          }

          JsonNode layerNode = layerWrapperNode.elements().next();
          JsonNode weightInit = layerNode.get(
              "weightInit"); //Should only have 1 element: "dense", "output", etc
          JsonNode distribution = layerNode.get("dist");

          Distribution dist = null;
          if (distribution != null) {
            dist = mapper.treeToValue(distribution, Distribution.class);
          }

          if (weightInit != null) {
            final IWeightInit wi = WeightInit.valueOf(weightInit.asText())
                .getWeightInitFunction(dist);
            ((BaseLayerConfiguration) l).setWeightInit(wi);
          }
        }

      } catch (IOException e) {
        log.warn(
            "ILayer with null WeightInit detected: " + l.getLayerName() + ", could not parse JSON",
            e);
      }
    }
    return true;

  }

  /**
   * Object mapper for serialization of configurations
   *
   * @return
   */
  public static ObjectMapper mapperYaml() {
    return JsonMappers.getMapperYaml();
  }

  /**
   * Object mapper for serialization of configurations
   *
   * @return
   */
  public static ObjectMapper mapper() {
    return JsonMappers.getMapper();
  }

  public static NeuralNetConfiguration fromYaml(String input) {
    throw new RuntimeException("Needs fixing - not supported."); //TODO
  }


  /**
   * @return JSON representation of NN configuration
   */
  public String toYaml() {
    ObjectMapper mapper = NeuralNetConfiguration.mapperYaml();
    synchronized (mapper) {
      try {
        return mapper.writeValueAsString(this);
      } catch (com.fasterxml.jackson.core.JsonProcessingException e) {
        throw new RuntimeException(e);
      }
    }
  }

  /**
   * @return JSON representation of NN configuration
   */
  public String toJson() {
    ObjectMapper mapper = NeuralNetConfiguration.mapper();
    synchronized (mapper) {
      //JSON mappers are supposed to be thread safe: however, in practice they seem to miss fields occasionally
      //when writeValueAsString is used by multiple threads. This results in invalid JSON. See issue #3243
      try {
        return mapper.writeValueAsString(this);
      } catch (com.fasterxml.jackson.core.JsonProcessingException e) {
        log.error(e.getMessage());
        throw new RuntimeException(e);
      }
    }
  }

  @Override
  public String toString() {
    return toJson();
  }

  @Override
  public NeuralNetConfiguration clone() {
    NeuralNetConfiguration clone;
    clone = (NeuralNetConfiguration) super.clone();
    clone.stepFunction = clone.stepFunction.clone();
    clone.netWideVariables = new ArrayList<>(netWideVariables);
    clone.getInnerConfigurations().addAll(innerConfigurations);

    if (clone.getInputPreProcessors() != null) {
      Map<Integer, InputPreProcessor> map = new HashMap<>();
      for (Map.Entry<Integer, InputPreProcessor> entry : clone.getInputPreProcessors().entrySet()) {
        map.put(entry.getKey(), entry.getValue().clone());
      }
      clone.getInputPreProcessors().clear();
      clone.getInputPreProcessors().putAll(map);
    }

    clone.setInferenceWorkspaceMode(this.inferenceWorkspaceMode);
    clone.setTrainingWorkspaceMode(this.trainingWorkspaceMode);
    clone.setCacheMode(this.cacheMode);
    clone.setValidateOutputLayerConfig(this.validateOutputLayerConfig);
    clone.setDataType(this.dataType);

    return clone;

  }

  /**
   *
   */
  @Override
  public void init() {
    if(initCalled) return;
    initCalled=true;

    /**
     * Run init() for each layer
     */

    getNetConfigurations().stream().forEach( conf -> {
      conf.init(); //do not call on self
    }); //call init on all embedded net configurations
      innerConfigurations.add(0, this); //put this configuration at first place

    /**
     * Inherit network wide configuration setting to those layer configurations
     * that do not have an individual setting (nor a default)
     */
    for(LayerConfiguration lconf : this.getFlattenedLayerConfigurations()) {
      if(lconf.getActivationFn() == null ) lconf.setActivationFn(this.getActivation());
      if(lconf.getIUpdater() == null ) lconf.setIUpdater( this.getUpdater() );
      if(lconf.getIDropout() == null ) lconf.setIDropout( this.getIdropOut() );
      if(lconf.getWeightNoise() == null ) lconf.setWeightNoise( this.getWeightNoise());

      // ... maybe more to set here ...
      if(lconf instanceof BaseLayerConfiguration ) { // then we can set some additional config settings
        BaseLayerConfiguration bconf = (BaseLayerConfiguration) lconf;
        if(bconf.getBiasUpdater() == null) bconf.setBiasUpdater(this.getBiasUpdater());
        if(bconf.getGradientNormalization() == null) bconf.setGradientNormalization(this.getGradientNormalization());
        // ... maybe more to set here ...
      }
    }



    getLayerConfigurations().stream().forEach( lconf -> lconf.setNetConfiguration(this)); //set this as net config for all layers (defined in here, not stacked


    //Validate BackpropType setting
    if ((tbpttBackLength != DEFAULT_TBPTT_LENGTH || tbpttFwdLength != DEFAULT_TBPTT_LENGTH)
        && backpropType != BackpropType.TruncatedBPTT) {
      log.warn("Truncated backpropagation through time lengths have been configured with values "
          + tbpttFwdLength
          + " and " + tbpttBackLength + " but backprop type is set to " + backpropType
          + ". TBPTT configuration" +
          " settings will only take effect if backprop type is set to BackpropType.TruncatedBPTT");
    }

    if (backpropType == BackpropType.TruncatedBPTT && validateTbpttConfig) {
      //Check for invalid combination - tbptt plus LastTimeStepLayer or
      for (int i = 0; i < getFlattenedLayerConfigurations().size(); i++) {
        LayerConfiguration l = getFlattenedLayerConfigurations().get(i);
        if (l instanceof LastTimeStep || l instanceof GlobalPoolingLayer) {
          throw new IllegalStateException(
              "Invalid network configuration detected: Truncated backpropagation through time (TBPTT)"
                  +
                  " cannot be used with layer " + i + " of type " + l.getClass().getName()
                  + ": TBPTT is incompatible with this layer type (which is designed " +
                  "to process entire sequences at once, and does support the type of sequence segments that TPBTT uses).\n"
                  +
                  "This check can be disabled using validateTbpttConfig(false) but this is not recommended.");
        }
      }
    }

    if (inputType == null && inputPreProcessors.get(0) == null) {
      //User hasn't set the InputType. Sometimes we can infer it...
      // For example, Dense/RNN layers, where preprocessor isn't set -> user is *probably* going to feed in
      // standard feedforward or RNN data
      //This isn't the most elegant implementation, but should avoid breaking backward compatibility here
      //Can't infer InputType for CNN layers, however (don't know image dimensions/depth)
      LayerConfiguration firstLayer = getFlattenedLayerConfigurations().get(0);
      if (firstLayer instanceof BaseRecurrentLayer) {
        BaseRecurrentLayer brl = (BaseRecurrentLayer) firstLayer;
        val nIn = brl.getNIn();
        if (nIn > 0) {
          inputType = InputType.recurrent(nIn, brl.getRnnDataFormat());
        }
      } else if (firstLayer instanceof DenseLayer || firstLayer instanceof EmbeddingLayer
          || firstLayer instanceof OutputLayer) {
        //Can't just use "instanceof FeedForwardLayer" here. ConvolutionLayer is also a FeedForwardLayer
        FeedForwardLayer ffl = (FeedForwardLayer) firstLayer;
        val nIn = ffl.getNIn();
        if (nIn > 0) {
          inputType = InputType.feedForward(nIn);
        }
      }
    }

    //Add preprocessors and set nIns, if InputType has been set
    // Builder.inputType field can be set in 1 of 4 ways:
    // 1. User calls setInputType directly
    // 2. Via ConvolutionLayerSetup -> internally calls setInputType(InputType.convolutional(...))
    // 3. Via the above code: i.e., assume input is as expected  by the RNN or dense layer -> sets the inputType field
    if(inputPreProcessors == null) {
      inputPreProcessors = new HashMap<>();
    }
    if (inputType != null) {
      InputType currentInputType = inputType;
      for (int i = 0; i < getFlattenedLayerConfigurations().size(); i++) {
        LayerConfiguration l = getFlattenedLayerConfigurations().get(i);
        if (inputPreProcessors.get(i) == null) {
          //Don't override preprocessor setting, but set preprocessor if required...
          @NonNull
          InputPreProcessor inputPreProcessor = l.getPreProcessorForInputType(currentInputType);
          if (inputPreProcessor != null) {
            inputPreProcessors.put(i, inputPreProcessor);
          }
        }

        InputPreProcessor inputPreProcessor = inputPreProcessors.get(i);
        if (inputPreProcessor != null) {
          currentInputType = inputPreProcessor.getOutputType(currentInputType);
        }
        if (i > 0) {
          LayerConfiguration layer = getFlattenedLayerConfigurations().get(i - 1);
          //convolution 1d is an edge case where it has rnn input type but the filters
          //should be the output
          if (layer instanceof Convolution1DLayer) {
            if (l instanceof DenseLayer && inputType instanceof InputType.InputTypeRecurrent) {
              FeedForwardLayer feedForwardLayer = (FeedForwardLayer) l;
              if (inputType instanceof InputType.InputTypeRecurrent) {
                InputType.InputTypeRecurrent recurrent = (InputType.InputTypeRecurrent) inputType;
                feedForwardLayer.setNIn(recurrent.getTimeSeriesLength());
              }
            } else {
              l.setNIn(currentInputType,
                  overrideNinUponBuild); //Don't override the nIn setting, if it's manually set by the user
            }
          } else {
            l.setNIn(currentInputType,
                overrideNinUponBuild); //Don't override the nIn setting, if it's manually set by the user
          }

        } else {
          l.setNIn(currentInputType,
              overrideNinUponBuild); //Don't override the nIn setting, if it's manually set by the user
        }

        currentInputType = l.getOutputType(i, currentInputType);
      }

    }

    Nd4j.getRandom().setSeed(getNetConfigurations().get(0).getSeed());

    //Validate output layer configuration
    if (isValidateOutputLayerConfig()) {
      //Validate output layer configurations...
      for (LayerConfiguration n : getFlattenedLayerConfigurations()) {
        OutputLayerUtil.validateOutputLayer(n.getLayerName(), n); //No-op for non output/loss layers
      }
    }
  }

  public InputPreProcessor getInputPreProcess(int curr) {
    return inputPreProcessors.get(curr);
  }

  /**
   * Get a {@link MemoryReport} for the given NeuralNetConfiguration. This is used to estimate the
   * memory requirements for the given network configuration and input
   *
   * @param inputType Input types for the network
   * @return Memory report for the network
   */
  public NetworkMemoryReport getMemoryReport(InputType inputType) {

    Map<String, MemoryReport> memoryReportMap = new LinkedHashMap<>();
    int nLayers = getFlattenedLayerConfigurations().size();
    for (int i = 0; i < nLayers; i++) {
      String layerName = getFlattenedLayerConfigurations().get(i).getLayerName();
      if (layerName == null) {
        layerName = String.valueOf(i);
      }

      //Pass input type through preprocessor, if necessary
      InputPreProcessor preproc = getInputPreProcess(i);
      //TODO memory requirements for preprocessor
      if (preproc != null) {
        inputType = preproc.getOutputType(inputType);
      }

      LayerMemoryReport report = getFlattenedLayerConfigurations().get(i).getMemoryReport(inputType);
      memoryReportMap.put(layerName, report);

      inputType = getFlattenedLayerConfigurations().get(i).getOutputType(i, inputType);
    }

    return new NetworkMemoryReport(memoryReportMap, NeuralNetConfiguration.class,
        "MultiLayerNetwork", inputType);
  }

  /**
   * For the given input shape/type for the network, return a list of activation sizes for each
   * layer in the network.<br> i.e., list.get(i) is the output activation sizes for layer i
   *
   * @param inputType Input type for the network
   * @return A lits of activation types for the network, indexed by layer number
   */
  public List<InputType> getLayerActivationTypes(@NonNull InputType inputType) {
    List<InputType> out = new ArrayList<>();
    int nLayers = getFlattenedLayerConfigurations().size();
    for (int i = 0; i < nLayers; i++) {
      InputPreProcessor preproc = getInputPreProcess(i);
      if (preproc != null) {
        inputType = preproc.getOutputType(inputType);
      }

      inputType = getFlattenedLayerConfigurations().get(i).getOutputType(i, inputType);
      out.add(inputType);
    }
    return out;
  }


  public List<String> netWideVariables() {

    return netWideVariables;
  }

  public List<String> netWideVariables(boolean copy) {
    if (copy) {
      return netWideVariables();
    }
    return netWideVariables;
  }

  public void addNetWideVariable(String variable) {
    if (!netWideVariables.contains(variable)) {
      netWideVariables.add(variable);
      log.trace("Adding neural network wide variable '{}' to the list of variables. New length is {}.", variable, netWideVariables.size());
    }
    log.trace("Skipped adding neural network wide variable '{}' to the list of variables. It was already present. Length remains {}.", variable, netWideVariables.size());
  }

  public void clearNetWideVariable() {

    netWideVariables.clear();
    log.trace("Adding neural network wide variables have been cleared. New length is {}.", netWideVariables.size());
  }



  /**
   * From the list of layers and neural net configurations, only return the Layer Configurations that
   * are defined in this neural network (it does not include embedded neural network configuration
   * layers)
   * @return list with layer configurations
   */
  public List<LayerConfiguration> getLayerConfigurations() {
    return innerConfigurations.stream()
        .filter(obj -> (obj instanceof LayerConfiguration))
        .map( obj -> (LayerConfiguration)obj )
        .collect( Collectors.toList());
  }

  /**
   * From the list of layers and neural net configurations, only return the neural net configurations
   * @return list with neural net configurations
   */
  public List<NeuralNetConfiguration> getNetConfigurations() {
    return innerConfigurations.stream()
        .filter(obj -> (obj instanceof NeuralNetConfiguration))
        .map( obj -> (NeuralNetConfiguration)obj )
        .collect( Collectors.toList());
  }

  /**
   * From the list of layer configurations and inner neural net configurations, create a single,
   * flattened list of layer configurations with inheritance parameters resolved
   *
   * @return list of layer configurations
   */
  public List<LayerConfiguration> getFlattenedLayerConfigurations(NeuralNetConfiguration conf) {
    List<LayerConfiguration> ret = new ArrayList<>(); //create the final return list
    //When properly initialized, _this_ configuration is set first in the list, however we
    //can find cases where this is not true, thus the first configuration is another net or layer configuration
    //and should not be skipped. In essence, skip first configuration if that is "this".
    int iSkip = 0;
    if(conf.getInnerConfigurations().size()>0 && conf.getInnerConfigurations().get(0).equals(this)) { iSkip=1;}
    conf.getInnerConfigurations().stream().skip(iSkip)
            .forEach(obj -> {
              //if Layer Config, include in list and inherit parameters from this conf
              //else if neural net configuration, call self recursively to resolve layer configurations
              if (obj instanceof LayerConfiguration) {
                ((LayerConfiguration) obj).setNetConfiguration(conf);
                ret.add((LayerConfiguration) obj);
                } else if (obj instanceof NeuralNetConfiguration)
                ret.addAll(getFlattenedLayerConfigurations(
                        (NeuralNetConfiguration) obj));
              else {
                log.error(
                        "The list of layers and neural network configurations does contain an object of {}. Element will be ignored.",
                        obj.getClass().getSimpleName());
              }
            });
    return ret;
  }

  /**
   * Sames as {@link #getFlattenedLayerConfigurations(NeuralNetConfiguration)}, but uses this configurations
   * list of configurations
   * @return list of layer configurations
   */
  public List<LayerConfiguration> getFlattenedLayerConfigurations() {
    return getFlattenedLayerConfigurations(this);
  }

  /**
   * Add a new layer to the first position
   * @param layer configuration
   */
  public void setLayer(@NonNull LayerConfiguration layer) {
    innerConfigurations.add(0, layer);
  }

  @Deprecated
  public LayerConfiguration getConf(int index) {
    return getFlattenedLayerConfigurations().get(index);
  }

  /**
   * Deprecated, do not use. Workaround for old tests
   * and getFlattenedLayerConfigurations().get(0);
   * @return
   */
  @Deprecated @JsonIgnore
  public LayerConfiguration getFirstLayer() {
    log.warn("This getFirstLayer method is an ugly workaround and will be removed.");
    return getFlattenedLayerConfigurations().get(0);
  }

  public static abstract class NeuralNetConfigurationBuilder<C extends NeuralNetConfiguration,
      B extends NeuralNetConfiguration.NeuralNetConfigurationBuilder<C, B>> extends
      NeuralNetBaseBuilderConfigurationBuilder<C, B> {

    public ComputationGraphConfiguration.GraphBuilder graphBuilder() {
      return new ComputationGraphConfiguration.GraphBuilder(this);
    }

    public NeuralNetConfigurationBuilder clone() {
      try {
        return (NeuralNetConfigurationBuilder) super.clone();
      } catch(CloneNotSupportedException ex) {
        throw new RuntimeException(ex);
      }
    }

  }

  public IModel getNeuralNet() {
    return net;
  }

  public void setNeuralNet(IModel model) {
    this.net = model;
  }

}
