/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.nn.conf.layers.wrapper;

import java.util.List;
import lombok.Data;
import lombok.EqualsAndHashCode;
import org.deeplearning4j.nn.api.ParamInitializer;
import org.deeplearning4j.nn.conf.GradientNormalization;
import org.deeplearning4j.nn.conf.InputPreProcessor;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.dropout.IDropout;
import org.deeplearning4j.nn.conf.inputs.InputType;
import org.deeplearning4j.nn.conf.layers.LayerConfiguration;
import org.deeplearning4j.nn.conf.memory.LayerMemoryReport;
import org.deeplearning4j.nn.conf.weightnoise.IWeightNoise;
import org.deeplearning4j.nn.params.WrapperLayerParamInitializer;
import org.nd4j.linalg.activations.IActivation;
import org.nd4j.linalg.learning.config.IUpdater;
import org.nd4j.linalg.learning.regularization.Regularization;

@Data
@EqualsAndHashCode(callSuper = false)
public abstract class BaseWrapperLayerConfiguration extends LayerConfiguration {

    /**
     * The configuration to of another layer to wrap
     */
  protected LayerConfiguration underlying;

  protected BaseWrapperLayerConfiguration(Builder builder) {
    super(builder);
  }

  protected BaseWrapperLayerConfiguration() {
  }

  public BaseWrapperLayerConfiguration(LayerConfiguration underlying) {
    this.underlying = underlying;
    this.setNetConfiguration(underlying.getNetConfiguration());
  }

  /**
   * Set the net configuration for this configuration as well as for the underlying layer (if not
   * null there)
   *
   * @param netConfiguration the neural net configuration
   */
  @Override
  public void setNetConfiguration(NeuralNetConfiguration netConfiguration) {
    super.setNetConfiguration(netConfiguration);
    if (underlying.getNetConfiguration() == null) {
        underlying.setNetConfiguration(
          netConfiguration); //also set netconf for underlying if not set
    }
  }

  /**
   * @return
   */
  @Override
  public IActivation getActivationFn() {
    return underlying.getActivationFn();
  }

  /**
   * @return
   */
  @Override
  public IDropout getIDropout() {
    return underlying.getIDropout();
  }

  /**
   * @param activationFn
   */
  @Override
  public void setActivationFn(IActivation activationFn) {
    underlying.setActivationFn(activationFn);
  }

  /**
   * @param iDropout
   */
  @Override
  public void setIDropout(IDropout iDropout) {
    underlying.setIDropout(iDropout);
  }

  /**
   * @param weightNoise
   */
  @Override
  public void setWeightNoise(IWeightNoise weightNoise) {
    underlying.setWeightNoise(weightNoise);
  }

  /**
   * @param s
   */
  @Override
  public void addVariable(String s) {
    underlying.addVariable(s);
  }

  /**
   * Get the updater for the given parameter. Typically, the same updater will be used for all
   * parameters, but this is not necessarily the case
   *
   * @param paramName Parameter name
   * @return IUpdater for the parameter
   */
  @Override
  public IUpdater getUpdaterByParam(String paramName) {
    return underlying.getUpdaterByParam(paramName);
  }

  /**
   * @param iUpdater
   */
  @Override
  public void setIUpdater(IUpdater iUpdater) {
    underlying.setIUpdater(iUpdater);
  }

  /**
   * @return
   */
  @Override
  public IUpdater getIUpdater() {
    return underlying.getIUpdater();
  }

  @Override
  public ParamInitializer initializer() {
    return WrapperLayerParamInitializer.getInstance();
  }

  @Override
  public InputType getOutputType(int layerIndex, InputType inputType) {
    return underlying.getOutputType(layerIndex, inputType);
  }

  @Override
  public void setNIn(InputType inputType, boolean override) {
    underlying.setNIn(inputType, override);
  }

  @Override
  public InputPreProcessor getPreProcessorForInputType(InputType inputType) {
    return underlying.getPreProcessorForInputType(inputType);
  }

  @Override
  public List<Regularization> getRegularizationByParam(String paramName) {
    return underlying.getRegularizationByParam(paramName);
  }

  @Override
  public boolean isPretrainParam(String paramName) {
    return underlying.isPretrainParam(paramName);
  }

  @Override
  public LayerMemoryReport getMemoryReport(InputType inputType) {
    return underlying.getMemoryReport(inputType);
  }

  @Override
  public void setLayerName(String layerName) {
    super.setLayerName(layerName);
    if (underlying != null) {
      //May be null at some points during JSON deserialization
      underlying.setLayerName(layerName);
    }
  }

}
