/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.nn.layers;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import lombok.*;
import net.brutex.ai.dnn.api.IModel;
import org.deeplearning4j.nn.api.ITraininableLayerConfiguration;
import org.deeplearning4j.nn.api.Layer;
import org.deeplearning4j.nn.api.MaskState;
import org.deeplearning4j.nn.api.layers.LayerConstraint;
import org.deeplearning4j.nn.conf.CacheMode;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.layers.LayerConfiguration;
import org.deeplearning4j.nn.gradient.Gradient;
import org.deeplearning4j.nn.workspace.ArrayType;
import org.deeplearning4j.nn.workspace.LayerWorkspaceMgr;
import org.deeplearning4j.optimize.api.ConvexOptimizer;
import org.deeplearning4j.optimize.api.TrainingListener;
import org.nd4j.common.primitives.Pair;
import org.nd4j.evaluation.IEvaluation;
import org.nd4j.linalg.api.buffer.DataType;
import org.nd4j.linalg.api.ndarray.INDArray;
import org.nd4j.linalg.dataset.api.DataSet;
import org.nd4j.linalg.dataset.api.MultiDataSet;
import org.nd4j.linalg.dataset.api.iterator.DataSetIterator;
import org.nd4j.linalg.dataset.api.iterator.MultiDataSetIterator;

/** A layer with input and output, no parameters or gradients */
@NoArgsConstructor(force = true)
public abstract class AbstractLayer<LayerConf_T extends LayerConfiguration> implements Layer {

  private final @Getter List<String> variables = new ArrayList<>();

  @Getter
  @Setter(AccessLevel.MODULE)
  protected INDArray
      input; // TODO: this should be private, but too much code is still accessing input directly.

  protected INDArray preOutput;
  /** The typed {@link LayerConfiguration}. */
  @Getter @NonNull protected LayerConf_T layerConfiguration;

  protected boolean dropoutApplied = false;

  @Getter @Setter @NonNull
  protected Collection<TrainingListener> trainingListeners = new ArrayList<>();

  protected @Getter @Setter int index = 0;
  protected @Getter @Setter INDArray maskArray;
  protected @Getter @Setter MaskState maskState;
  protected CacheMode cacheMode = CacheMode.NONE;
  protected boolean inputModificationAllowed = false;
  protected DataType dataType;
  protected @Getter @Setter int iterationCount;
  protected @Getter @Setter int epochCount;
  private @Getter @Setter IModel net;

  @Getter @Setter @NonNull private NeuralNetConfiguration netConfiguration;

  public AbstractLayer(@NonNull LayerConfiguration layerConf, @NonNull DataType dataType) {
    //noinspection unchecked
    this.layerConfiguration = (LayerConf_T) layerConf;
    this.netConfiguration = layerConfiguration.getNetConfiguration();

    if (layerConfiguration.getNetConfiguration() != null) {
      cacheMode = layerConfiguration.getNetConfiguration().getCacheMode();
    }
    this.dataType = dataType;
    if (layerConfiguration.getNetConfiguration() == null) {
      throw new RuntimeException("You cannot create a layer from a layer configuration, that is not part of any neural network configuration.");
      }
    this.net = layerConfiguration.getNetConfiguration().getNet();
  }

  public void addTrainingListeners(TrainingListener... listeners) {
    if(listeners != null)
    trainingListeners.addAll(List.of(listeners));
  }

  public void addTrainingListeners(Collection<TrainingListener> listeners) {
    if(listeners != null)
    trainingListeners.addAll(listeners);
  }

  @Override
  public INDArray activate(INDArray input, boolean training, LayerWorkspaceMgr workspaceMgr) {
    setInput(input, workspaceMgr);
    return activate(training, workspaceMgr);
  }

  /**
   * Creates and returns a copy of this object.
   *
   * @return a clone of this instance.
   * @throws CloneNotSupportedException if the object's class does not support the {@code Cloneable}
   *     interface. Subclasses that override the {@code clone} method can also throw this exception
   *     to indicate that an instance cannot be cloned.
   * @see Cloneable
   */
  @Override
  protected Object clone() throws CloneNotSupportedException {
    return super.clone();
  }

  /**
   * This method returns updater state (if applicable), null otherwise
   *
   * @return
   */
  @Override
  public INDArray updaterState() {
    return null;
  }

  /**
   * This method returns Optimizer used for training
   *
   * @return
   */
  @Override
  public ConvexOptimizer getOptimizer() {
    return null;
  }

  /**
   * This method fits model with a given DataSet
   *
   * @param dataSet
   */
  @Override
  public void fit(DataSet dataSet) {}

  /**
   * This method fits model with a given MultiDataSet
   *
   * @param dataSet
   */
  @Override
  public void fit(MultiDataSet dataSet) {}

  /**
   * This method fits model with a given DataSetIterator
   *
   * @param iterator
   */
  @Override
  public void fit(DataSetIterator iterator) {}

  /**
   * This method fits model with a given MultiDataSetIterator
   *
   * @param iterator
   */
  @Override
  public void fit(MultiDataSetIterator iterator) {}

  /**
   * This method executes evaluation of the model against given iterator and evaluation
   * implementations
   *
   * @param iterator
   * @param evaluations
   */
  @Override
  public <T extends IEvaluation> T[] doEvaluation(DataSetIterator iterator, T... evaluations) {
    return null;
  }

  /**
   * This method executes evaluation of the model against given iterator and evaluation
   * implementations
   *
   * @param iterator
   * @param evaluations
   */
  @Override
  public <T extends IEvaluation> T[] doEvaluation(MultiDataSetIterator iterator, T... evaluations) {
    return null;
  }

  /** Init the model */
  @Override
  public void init() {}

  /**
   * Update layer weights and biases with gradient change
   *
   * @param gradient
   */
  @Override
  public void update(Gradient gradient) {}

  /**
   * Perform one update applying the gradient
   *
   * @param gradient the gradient to apply
   * @param paramType
   */
  @Override
  public void update(INDArray gradient, String paramType) {}

  /**
   * Update the score
   *
   * @param workspaceMgr
   */
  @Override
  public void computeGradientAndScore(LayerWorkspaceMgr workspaceMgr) {}

  /**
   * the number of parameters for the model
   *
   * @param backwards
   * @return the number of parameters for the model
   */
  @Override
  public long numParams(boolean backwards) {
    return 0;
  }

  @Override
  public void setParam(String s, INDArray array) {}

  /**
   * Get a parameter array for a given parameter type key
   *
   * @param param the key of the parameter
   * @return ndarray of parameters
   */
  @Override
  public INDArray getParam(String param) {
    return null;
  }

  /**
   * Set the initial parameters array as a view of the full (backprop) network parameters NOTE: this
   * is intended to be used internally in MultiLayerNetwork and ComputationGraph, not by users.
   *
   * @param params a 1 x nParams row vector that is a view of the larger (MLN/CG) parameters array
   */
  @Override
  public void setParamsViewArray(INDArray params) {}

  /**
   * Set the gradients array as a view of the full (backprop) network parameters NOTE: this is
   * intended to be used internally in MultiLayerNetwork and ComputationGraph, not by users.
   *
   * @param gradients a 1 x nParams row vector that is a view of the larger (MLN/CG) gradients array
   */
  @Override
  public void setBackpropGradientsViewArray(INDArray gradients) {}

  /**
   * The current inputs batch size
   *
   * @return the current inputs batch size
   */
  @Override
  public int batchSize() {
    return 0;
  }

  /**
   * The input/feature matrix for the model
   *
   * @return the input/feature matrix for the model
   */
  @Override
  public INDArray input() {
    return this.input;
  }

  /** */
  @Override
  public void close() {}

  /**
   * Calculate the gradient relative to the error in the next layer
   *
   * @param epsilon w^(L+1)*delta^(L+1). Or, equiv: dC/da, i.e., (dC/dz)*(dz/da) = dC/da, where C is
   *     cost function a=sigma(z) is activation.
   * @param workspaceMgr Workspace manager
   * @return Pair<Gradient, INDArray> where Gradient is gradient for this layer, INDArray is epsilon
   *     (activation gradient) needed by next layer, but before element-wise multiply by
   *     sigmaPrime(z). So for standard feed-forward layer, if this layer is L, then
   *     return.getSecond() == dL/dIn = (w^(L)*(delta^(L))^T)^T. Note that the returned array should
   *     be placed in the {@link ArrayType#ACTIVATION_GRAD} workspace via the workspace manager
   */
  @Override
  public Pair<Gradient, INDArray> backpropGradient(
      INDArray epsilon, LayerWorkspaceMgr workspaceMgr) {
    return null;
  }

  /**
   * Returns true if the layer can be trained in an unsupervised/pretrain manner (AE, VAE, etc)
   *
   * @return true if the layer can be pretrained (using fit(INDArray), false otherwise
   */
  @Override
  public boolean isPretrainLayer() {
    return false;
  }

  /** */
  @Override
  public void clearNoiseWeightParams() {}

  public List<String> getVariables(boolean copy) {
    if (copy) {
      return new ArrayList<>(getVariables());
    }
    return variables;
  }

  public void addVariable(String variable) {
    if (!variables.contains(variable)) {
      variables.add(variable);
    }
  }

  public void setLayerConfiguration(LayerConfiguration layerConfiguration) {
    //noinspection unchecked
    this.layerConfiguration = (LayerConf_T) layerConfiguration;
  }

  @Override
  public void setCacheMode(CacheMode mode) {
    if (mode == null) {
      mode = CacheMode.NONE;
    }

    this.cacheMode = mode;
  }

  public LayerConf_T getTypedLayerConfiguration() {
    return this.layerConfiguration;
  }

  @Override
  public ITraininableLayerConfiguration getTrainingConfig() {
    return (ITraininableLayerConfiguration) getTypedLayerConfiguration();
  }

  protected String layerId() {
    String name = this.layerConfiguration.getLayerName();
    return "(layer name: "
        + (name == null ? "\"\"" : name)
        + ", layer index: "
        + index
        + ", layer type: "
        + getClass().getSimpleName()
        + ")";
  }

  @Override
  public void setInput(@NonNull INDArray input, LayerWorkspaceMgr workspaceMgr) {
    this.input = workspaceMgr.leverageTo(ArrayType.INPUT, input);
    dropoutApplied = false;
  }

  /**
   * Returns the parameters of the neural network as a flattened row vector
   *
   * @return the parameters of the neural network
   */
  @Override
  public INDArray getModelParams() {
    return null;
  }

  protected void setParams(INDArray params, char order) {
    throw new UnsupportedOperationException("Not supported");
  }

  /**
   * @return Number of parameters
   */
  @Override
  public long numParams() {
    return 0;
  }

  protected void applyMask(INDArray to) {
    to.muliColumnVector(maskArray.castTo(to.dataType()));
  }

  @Override
  public double calcRegularizationScore(boolean backpropParamsOnly) {
    return 0.0;
  }

  @Deprecated
  public void clear() {
    input = null;
    maskArray = null;
    maskState = null;
    if (getTypedLayerConfiguration().getIDropout() != null) {
      getTypedLayerConfiguration().getIDropout().clear();
    }
  }

  protected void applyDropOutIfNecessary(boolean training, LayerWorkspaceMgr workspaceMgr) {
    if (training && !dropoutApplied && getTypedLayerConfiguration().getIDropout() != null) {
      INDArray result;
      if (inputModificationAllowed) {
        result = input;
      } else {
        result =
            workspaceMgr.createUninitialized(
                ArrayType.INPUT, input.dataType(), input.shape(), input.ordering());
      }

      input =
          getTypedLayerConfiguration()
              .getIDropout()
              .applyDropout(input, result, getIterationCount(), getEpochCount(), workspaceMgr);
      dropoutApplied = true;
    }
  }

  protected INDArray backpropDropOutIfPresent(INDArray epsilon) {
    if (getTypedLayerConfiguration().getIDropout() != null) {
      getTypedLayerConfiguration()
          .getIDropout()
          .backprop(epsilon, epsilon, getIterationCount(), getEpochCount());
    }
    return epsilon;
  }

  @Override
  public Type type() {
    return Type.FEED_FORWARD;
  }

  public void fit(INDArray input, LayerWorkspaceMgr workspaceMgr) {
    throw new UnsupportedOperationException("Not supported");
  }

  public Pair<Gradient, Double> gradientAndScore() {
    return new Pair<>(gradient(), getScore());
  }

  @Override
  public int getInputMiniBatchSize() {
    return (int) input.size(0);
  }

  @Override
  public void setInputMiniBatchSize(int size) {}

  @Override
  public Pair<INDArray, MaskState> feedForwardMaskArray(
      INDArray maskArray, MaskState currentMaskState, int minibatchSize) {
    // Most layers: CNN, dense, activation, etc - set mask array, mask state and then leave the mask
    // unmodified

    this.maskArray = maskArray;
    this.maskState = currentMaskState;

    return new Pair<>(maskArray, currentMaskState);
  }

  public Gradient gradient() {
    throw new UnsupportedOperationException(
        "Not supported for this layer, or should be overridden for layers requiring it");
  }

  public void fit() {
    throw new UnsupportedOperationException(
        "Not supported for this layer, or should be overridden for layers requiring it");
  }

  public double getScore() {
    throw new UnsupportedOperationException(
        "Not supported for this layer, or should be overridden for layers requiring it");
  }

  public void applyConstraints(int iteration, int epoch) {
    if (getTypedLayerConfiguration().getConstraints() != null) {
      for (LayerConstraint lc : getTypedLayerConfiguration().getConstraints()) {
        lc.applyConstraint(this, iteration, epoch);
      }
    }
  }

  public void assertInputSet(boolean backprop) {
    if (input == null) {
      if (backprop) {
        throw new IllegalStateException(
            "Cannot perform backprop in layer "
                + getClass().getSimpleName()
                + ": layer input field is not set");
      } else {
        throw new IllegalStateException(
            "Cannot perform forward pass in layer "
                + getClass().getSimpleName()
                + ": layer input field is not set");
      }
    }
  }

  @Override
  public void allowInputModification(boolean allow) {
    inputModificationAllowed = allow;
  }

  @Override
  public LayerHelper getHelper() {
    // Layers with helpers should override this method!
    return null;
  }

  @Override
  public boolean updaterDivideByMinibatch(String paramName) {
    // Majority of params's gradients should be... Exception: batch norm mean/variance estimate
    return true;
  }

  /**
   * The AbstractLayer does not implement Params, ParamTable and GradientView. A RuntimeException
   * will be triggered when calling this.
   *
   * @return
   */
  @Override
  public Map<String, INDArray> getParamTable() {
    throw new RuntimeException("Not implemented");
  }

  /**
   * * The AbstractLayer does not implement Params, ParamTable and GradientView. A RuntimeException
   * * will be triggered when calling this.
   *
   * @param paramTable
   */
  @Override
  public void setParamTable(Map<String, INDArray> paramTable) {
    throw new RuntimeException("Not implemented");
  }

  /**
   * * The AbstractLayer does not implement Params, ParamTable and GradientView. A RuntimeException
   * * will be triggered when calling this.
   *
   * @param isBackprop
   * @return
   */
  @Override
  public Map<String, INDArray> getParamTable(boolean isBackprop) {
    throw new RuntimeException("Not implemented");
  }

  /**
   * * The AbstractLayer does not implement Params, ParamTable and GradientView. A RuntimeException
   * * will be triggered when calling this.
   *
   * @return 1d parameter vector
   */
  @Override
  public INDArray getParams() {
    //throw new RuntimeException("Not implemented");
    return null;
  }

  /**
   * Set the parameters for this model. This expects a linear ndarray which then be unpacked
   * internally relative to the expected ordering of the model
   *
   * @param params the parameters for the model
   */
  @Override
  public void setParams(INDArray params) {}

  /**
   * * The AbstractLayer does not implement Params, ParamTable and GradientView. A RuntimeException
   * * will be triggered when calling this.
   *
   * @return 1D gradients view array
   */
  @Override
  public INDArray getGradientsViewArray() {
    throw new RuntimeException("Not implemented");
  }
}
